/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2010, Openvox Technologies Corp.
 *
 * Bo Zhou<bo.zhou@openvox.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief codec_openvox.c - Openvox Vocallo codec module
 * 
 * \ingroup codecs
 */

/*** MODULEINFO
	<depend>openvox_transcode</depend>
 ***/

/*
 * some notes about Asterisk versions and its interfaces, versions may vary slightly, I tested
 * with latest of each release, feel free to adjust if you find something inaccurate
 * > 10602 (at this time trunk) defines new RTP pluggable architecture and some new config keywords
 * >= 10601 defines data as a union instead of a pointer inside an ast_frame (f.data.ptr)
 * >= 10600 defines new CLI macros and prototypes for the callbacks 
 * the asteris/version.h macro ASTERISK_VERSION_NUM is used to keep this source compatible with all versions
 * >= 10400 asterisk.h was introduced
 *    ast_frame now has flags member
 *    the transcoding interface changed a lot, mostly the way the private structure is allocated and used
 * */
#include "asterisk/version.h"

#if ASTERISK_VERSION_NUM == 010400
#undef ASTERISK_VERSION_NUM
#define ASTERISK_VERSION_NUM 10400
#define ARRAY_LEN(a) sizeof(a)/sizeof(a[0])
#endif

#if ASTERISK_VERSION_NUM >= 10400
#include "asterisk.h"
#endif

/* common Asterisk includes */
#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/translate.h"
#include "asterisk/alaw.h"
#include "asterisk/utils.h"
#include "asterisk/sched.h"
#include "asterisk/cli.h"
#include "asterisk/acl.h"
#include "asterisk/app.h"

#if ASTERISK_VERSION_NUM > 10602
#include "asterisk/rtp_engine.h"
#else
#include "asterisk/rtp.h"
#endif

/* Openvox transcoding library */
#include <opx_tc/opx_tc.h>

#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <errno.h>
#include <execinfo.h>

/* default to 1 until bug http://issues.asterisk.org/view.php?id=17092 is fixed in Asterisk */
static int g_attempt_unload_conflicting_translators = 0;

static uint8_t g_rtcp_enable = 0;
/* we've seen delays, crashes and audio drops related to non-DAHDI timing, do not allow other timings by default */
static int g_allow_nondahdi_timing = 0;
static int g_dahdi_timing = 0;

/* supported codecs */
typedef struct vocallo_codec_s {
	int codec_id;
	int astformat;
	int maxms;
} vocallo_codec_t;

vocallo_codec_t g_codec_map[] =
{
	{ OPXTC_CODEC_PCMU,    AST_FORMAT_ULAW    , 40 },
	{ OPXTC_CODEC_PCMA,    AST_FORMAT_ALAW    , 40 },
//	{ OPXTC_CODEC_L16_1,   AST_FORMAT_SLINEAR , 40 },
	{ OPXTC_CODEC_G729AB,  AST_FORMAT_G729A   , 40 },
	{ OPXTC_CODEC_G722,    AST_FORMAT_G722    , 20 },
	{ OPXTC_CODEC_G726_32, AST_FORMAT_G726    , 40 }, 
	{ OPXTC_CODEC_GSM_FR,  AST_FORMAT_GSM     , 20 },
	{ OPXTC_CODEC_G723,    AST_FORMAT_G723_1  , 30 },
#if 1
	{ OPXTC_CODEC_ILBC,    AST_FORMAT_ILBC    , 30 },
//	{ OPXTC_CODEC_SPEEX,   AST_FORMAT_SPEEX   , 30 },
#endif
	{ -1, -1 },
};

#define MAX_CODECS_LIST 64
#define MAX_CODEC_STR 32
static unsigned int g_codec_register_list_count = 0;
static unsigned int g_codec_noregister_list_count = 0;
static char g_codec_register_list[MAX_CODECS_LIST][MAX_CODEC_STR];
static char g_codec_noregister_list[MAX_CODECS_LIST][MAX_CODEC_STR];

#define OPENVOX_TRANSLATOR_BUFFER_SIZE 8000
#define FAKE_FRAME_IDLE 0
#define FAKE_FRAME_REQUESTED 1
#define FAKE_FRAME_PROVIDED 2
#define FAKE_FRAME_TERMINATED 3 

#define OPENVOX_DEFAULT_SAMPLES 80*3
#define OPENVOX_DEFAULT_SAMPLING_RATE 80
#define OPENVOX_DEFAULT_RTP_STACK  "asterisk" 
#define OPENVOX_DEFAULT_UDP_PORT 15000
#define OPENVOX_TRANSCODE_CONFIG "openvox_codec.conf" 

#define OPX_SESS_WARN_TX(...) if (sess->txwarn > 10) { \
					sess->txwarn = sess->txwarn > 1000 ? 0 : sess->txwarn; \
					sess->txwarn++; \
				} else { \
					ast_log(LOG_WARNING, __VA_ARGS__); \
					sess->txwarn++; \
				}

#define OPX_SESS_WARN_RX(...) if (sess->rxwarn > 10) {  \
					sess->rxwarn = sess->rxwarn > 1000 ? 0 : sess->rxwarn; 		\
					sess->rxwarn++; \
				} else { \
					ast_log(LOG_WARNING, __VA_ARGS__); \
					sess->rxwarn++; \
				}

/*! transcoder init configuration */
static opxtc_init_cfg_t g_init_cfg;

/* \brief openvox transcoding session private structure */
#if ASTERISK_VERSION_NUM >= 10400
#define asterisk_transcoding_pvt ast_trans_pvt
#define openvox_transcoding_session openvox_transcoding_session
#define openvox_get_pvt(pvt) (pvt)->pvt
#else
#define asterisk_transcoding_pvt ast_translator_pvt
#define openvox_transcoding_session ast_translator_pvt
#define openvox_get_pvt(pvt) (pvt)
#endif

struct openvox_rtp_handle_ast {
	void *ast_rtp; /* The asterisk RTP handle */
	unsigned int flags;
	ast_mutex_t rxlock;
	ast_mutex_t txlock;
};

struct openvox_transcoding_session {
	struct asterisk_transcoding_pvt *owner;
	int fakestate;
	opxtc_codec_request_t codec_request;
	opxtc_codec_reply_t codec_reply;
	opxtc_rtp_stats_t rtp_stats;

	/* tx and rx rtp streams */
	struct openvox_rtp_handle_ast *txrtp;
	struct openvox_rtp_handle_ast *rxrtp;

	/* unique session id */
	unsigned long long sessid;
	
	/* tx calls and rx calls counters */
	unsigned long long txcount;
	unsigned long long rxcount;
	unsigned long long rxcallcount;
	unsigned long long rxnullcount;

	unsigned short txavgtime;
	unsigned short rxavgtime;
	unsigned short rxcallavgtime;

	struct timeval lasttxtime;
	struct timeval lastrxtime;
	struct timeval lastrxcalltime;

	/* seq counters */
  	long lastrxseqno;
	long lostrxpackets;
	/* warning counters */
	unsigned short rxwarn;
	unsigned short txwarn;
#if ASTERISK_VERSION_NUM < 10400
	/* this duplicates some info like openvox_transcoding_session which is the same as owner 
	 * but keeps the processing code across asterisk versions cleaner (less if ASTERISK_VERSION stuff) */
	int samples;
	int srcfmt;
	int dstfmt;
	struct openvox_transcoding_session *t;
	struct ast_frame f;
#endif
	AST_LIST_ENTRY(openvox_transcoding_session) entry;
};

/* \brief simple container to list all registered asterisk translators in this openvox module */
struct openvox_translator {
	struct ast_translator t;
	AST_LIST_ENTRY(openvox_translator) entry;
};

/* \brief global scheduling context */
static struct sched_context *g_sched = NULL;

/* \brief translators active flag: 1, active; 0, deactive */
//static char active = 1;

/* \brief global session id, the list of openvox translator sessions lock protects the global next id as well */
static unsigned long long g_next_session_id = 0;
static unsigned long long g_total_sessions  = 0;

/* \brief list of openvox translators, populated on module loading depending on which translators are available or configured */
static AST_LIST_HEAD_STATIC(g_openvox_translators, openvox_translator);

/* \brief list of current openvox transcoding sessions */
static AST_LIST_HEAD_STATIC(g_openvox_sessions, openvox_transcoding_session);

/* \brief Module description */
static char g_openvox_module_description[] = "Openvox Codec Translator";

/* \brief cli handler to show openvox transcoding usage */
#if ASTERISK_VERSION_NUM >= 10600
static char *handle_cli_openvox_show_translators(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_openvox_show_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_openvox_flush_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_openvox_show_rtp_statistics(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_openvox_show_license(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_openvox_show_rtcp(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_openvox_set_rtcp(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
#else
int handle_cli_openvox_show_translators(int fd, int argc, char *argv[]);
int handle_cli_openvox_show_transcoding(int fd, int argc, char *argv[]);
int handle_cli_openvox_flush_transcoding(int fd, int argc, char *argv[]);
int handle_cli_openvox_show_rtp_statistics(int fd, int argc, char *argv[]);
int handle_cli_openvox_show_license(int fd, int argc, char *argv[]);
int handle_cli_openvox_show_rtcp(int fd, int argc, char *argv[]);
int handle_cli_openvox_set_rtcp(int fd, int argc, char *argv[]);
#endif

static char g_cli_openvox_show_translators_usage[] = 
			"Usage: openvox show translators\n"
			"	Display Openvox registered translators\n";
static char g_cli_openvox_show_translators_description[] = "Display Openvox registered translators.";

static char g_cli_openvox_show_transcoding_usage[] = 
			"Usage: openvox show transcoding sessions\n"
			"		Display Openvox current transcoding sessions and statistics\n";
static char g_cli_openvox_show_transcoding_description[] = "Display Openvox transcoding utilization.";

static char g_cli_openvox_flush_transcoding_usage[] = 
			"Usage: openvox flush transcoding session\n"
			"		Flush all or only the provided openvox transcoding session RTP buffers\n";
static char g_cli_openvox_flush_transcoding_description[] = "Flush Openvox RTP transcoding buffers (be careful).";

static char g_cli_openvox_show_rtp_usage[] = 
			"Usage: openvox show rtp statistics\n"
			"		Display Openvox current RTP statistics per session\n";
static char g_cli_openvox_show_rtp_description[] = "Display Openvox RTP utilization.";

static char g_cli_openvox_show_license_usage[] = 
			"Usage: openvox show license\n"
			"		Display Openvox transcoding license information\n";
static char g_cli_openvox_show_license_description[] = "Display Openvox license information.";

static char g_cli_openvox_show_rtcp_usage[] = 
			"Usage: openvox show rtcp\n"
			"		Display Openvox transcoding rtcp setting.\n";
static char g_cli_openvox_show_rtcp_description[] = "Display Openvox transcoding rtcp setting.";

static char g_cli_openvox_set_rtcp_usage[] = 
			"Usage: openvox set rtcp {on|off}\n"
			"		Set Openvox transcoding rtcp.\n";
static char g_cli_openvox_set_rtcp_description[] = "Set Openvox transcoding rtcp.";

/* \brief array of cli entries to be registered with the Asterisk CLI core */
static struct ast_cli_entry g_openvox_cli[] = {
#if ASTERISK_VERSION_NUM >= 10600
	AST_CLI_DEFINE(handle_cli_openvox_show_translators, g_cli_openvox_show_translators_description),
	AST_CLI_DEFINE(handle_cli_openvox_show_transcoding, g_cli_openvox_show_transcoding_description),
	AST_CLI_DEFINE(handle_cli_openvox_flush_transcoding, g_cli_openvox_flush_transcoding_description),
	AST_CLI_DEFINE(handle_cli_openvox_show_rtp_statistics, g_cli_openvox_show_rtp_description),
	AST_CLI_DEFINE(handle_cli_openvox_show_license, g_cli_openvox_show_license_description),
	/*
	AST_CLI_DEFINE(handle_cli_openvox_show_rtcp, g_cli_openvox_show_rtcp_description),
	AST_CLI_DEFINE(handle_cli_openvox_set_rtcp, g_cli_openvox_set_rtcp_description)
	*/
#else
	{
		{ "openvox", "show", "translators", NULL },
		handle_cli_openvox_show_translators,
		g_cli_openvox_show_translators_description,
		g_cli_openvox_show_translators_usage, NULL, NULL, 
	},
	{
		{ "openvox", "show", "transcoding", "sessions", NULL },
		handle_cli_openvox_show_transcoding,
		g_cli_openvox_show_transcoding_description,
		g_cli_openvox_show_transcoding_usage, NULL, NULL, 
	},
	{
		{ "openvox", "flush", "transcoding", NULL },
		handle_cli_openvox_flush_transcoding,
		g_cli_openvox_flush_transcoding_description,
		g_cli_openvox_flush_transcoding_usage, NULL, NULL, 
	},
	{
		{ "openvox", "show", "rtp", "statistics", NULL },
		handle_cli_openvox_show_rtp_statistics,
		g_cli_openvox_show_rtp_description,
		g_cli_openvox_show_rtp_usage, NULL, NULL, 
	},
	{
		{ "openvox", "show", "license", NULL },
		handle_cli_openvox_show_license,
		g_cli_openvox_show_license_description,
		g_cli_openvox_show_license_usage, NULL, NULL, 
	},
	/*
	{
		{ "openvox", "show", "rtcp", NULL },
		handle_cli_openvox_show_rtcp,
		g_cli_openvox_show_rtcp_description,
		g_cli_openvox_show_rtcp_usage, NULL, NULL, 
	},
	{
		{ "openvox", "set", "rtcp", "on", NULL },
		handle_cli_openvox_set_rtcp,
		g_cli_openvox_set_rtcp_description,
		g_cli_openvox_set_rtcp_usage, NULL, NULL, 
	},
	{
		{ "openvox", "set", "rtcp", "off", NULL },
		handle_cli_openvox_set_rtcp,
		g_cli_openvox_set_rtcp_description,
		g_cli_openvox_set_rtcp_usage, NULL, NULL, 
	}
	*/
	
#endif
};

#if ASTERISK_VERSION_NUM < 10400
AST_MUTEX_DEFINE_STATIC(usemutex);
static int localusecnt = 0;
#define AST_MODULE_LOAD_DECLINE -1
#define AST_MODULE_LOAD_FAILURE -1
#define AST_MODULE_LOAD_SUCCESS 0
#define ast_free free
#define ast_calloc calloc
#define ARRAY_LEN(a) sizeof(a)/sizeof(a[0])
#define ast_vasprintf vasprintf
#endif

#if ASTERISK_VERSION_NUM > 10602
#define openvox_rtp_new(sess,localip) sess = ast_rtp_instance_new(OPENVOX_DEFAULT_RTP_STACK, g_sched, localip, NULL)
#define openvox_rtp_write ast_rtp_instance_write
#define openvox_rtp_read(instance) ast_rtp_instance_read(instance, 0)
#define openvox_rtp_get_local_address(sess, bindaddr) ast_rtp_instance_get_local_address(sess, bindaddr)
#define openvox_rtp_destroy ast_rtp_instance_destroy
#define openvox_rtp_set_remote_address ast_rtp_instance_set_remote_address
#else
#define openvox_rtp_new(sess,localip) sess = ast_rtp_new_with_bindaddr(g_sched, NULL, 0, 0, localip)
#define openvox_rtp_write ast_rtp_write
#define openvox_rtp_read ast_rtp_read
#define openvox_rtp_get_local_address(sess, bindaddr) ast_rtp_get_us(sess, bindaddr)
#define openvox_rtp_destroy ast_rtp_destroy
#define openvox_rtp_set_remote_address ast_rtp_set_peer
#endif

static void openvox_print_translators(int fd)
{
	int i = 0;
	struct openvox_translator *cur;
	AST_LIST_LOCK(&g_openvox_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_openvox_translators, cur, entry) {
#if 0
		ast_cli(fd, "%s Source Format: %s Destiny Format: %s\n", cur->t.name, ast_getformatname(1 << cur->t.srcfmt), 
				ast_getformatname(1 << cur->t.dstfmt));
#endif
		ast_cli(fd, "%s to %s\n", ast_getformatname(1 << cur->t.srcfmt), ast_getformatname(1 << cur->t.dstfmt));
		i++;
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_openvox_translators);
	ast_cli(fd, "Total translators: %d\n", i);
}

static void openvox_print_transcoding_sessions(int fd)
{
	struct openvox_transcoding_session *cur;
	AST_LIST_LOCK(&g_openvox_sessions);
	AST_LIST_TRAVERSE(&g_openvox_sessions, cur, entry) {
		ast_cli(fd, "[%03llu]%s to %s\tTx Frames: %llu Rx Frames: %llu Rx Nulls %llu Rx Calls %llu AvgRxCallTime %dms AvgRxTime %dms AvgTxTime %dms\n", cur->sessid, 
				ast_getformatname(1 << cur->owner->t->srcfmt), ast_getformatname(1 << cur->owner->t->dstfmt),
				cur->txcount, cur->rxcount, cur->rxnullcount, cur->rxcallcount, cur->rxcallavgtime, cur->rxavgtime, cur->txavgtime);
	}
	ast_cli(fd, "Total Sessions: %llu\n", g_total_sessions);
	AST_LIST_UNLOCK(&g_openvox_sessions);
}

#define BREAK_OR_CONTINUE(cursess, sessid) \
	if (sessid >= 0 && (cursess)->sessid == sessid) { \
		break; \
	} else { \
		continue; \
	}
static void openvox_flush_transcoding_sessions(int fd, int sessid)
{
	struct openvox_transcoding_session *cur;
	struct ast_frame *f;
	int res;

	if (sessid >= 0 ) {
		ast_cli(fd, "Flushing RTP for session %d only\n", sessid);
	} else {
		ast_log(LOG_WARNING, "Flushing all transcoding session buffers\n");
		ast_cli(fd, "Flushing RTP for all transcoding sessions\n");
	}

	AST_LIST_LOCK(&g_openvox_sessions);
	AST_LIST_TRAVERSE(&g_openvox_sessions, cur, entry) {

		if (sessid >= 0 && sessid != cur->sessid) {
			continue;
		}

		res = ast_mutex_lock(&cur->rxrtp->rxlock);
		if (res) {
			ast_log(LOG_ERROR, "Could not acquire rx lock during RTP flush in session %llu: (err %d)\n\n", cur->sessid, res);
			BREAK_OR_CONTINUE(cur, sessid);
		}

		do {
			f = openvox_rtp_read(cur->rxrtp->ast_rtp);
		} while (f && f->frametype != AST_FRAME_NULL);

		cur->rxcount = 0;
		cur->rxnullcount = 0;
		cur->rxcallavgtime = 0;
		cur->rxavgtime = 0;

		ast_mutex_unlock(&cur->rxrtp->rxlock);

		res = ast_mutex_lock(&cur->txrtp->txlock);
		if (res) {
			ast_log(LOG_ERROR, "Could not acquire tx lock during RTP flush in session %llu: (err %d)\n\n", cur->sessid, res);
			BREAK_OR_CONTINUE(cur, sessid);
		}

		cur->txcount = 0;
		cur->txavgtime = 0;

		ast_mutex_unlock(&cur->txrtp->txlock);

		ast_cli(fd, "[%03llu]%s to %s\tflushed\n", cur->sessid, ast_getformatname(1 << cur->owner->t->srcfmt), ast_getformatname(1 << cur->owner->t->dstfmt));

		BREAK_OR_CONTINUE(cur, sessid);
	}
	ast_cli(fd, "Total Sessions: %llu\n", g_total_sessions);
	AST_LIST_UNLOCK(&g_openvox_sessions);
}

static void openvox_print_rtp_statistics(int fd)
{
	struct openvox_transcoding_session *cur;
	AST_LIST_LOCK(&g_openvox_sessions);
	AST_LIST_TRAVERSE(&g_openvox_sessions, cur, entry) {
		ast_cli(fd, "[%03llu]%s to %s\tLost Frames: %li\n", cur->sessid, ast_getformatname(1 << cur->owner->t->srcfmt), 
				ast_getformatname(1 << cur->owner->t->dstfmt), cur->lostrxpackets);
	}
	AST_LIST_UNLOCK(&g_openvox_sessions);
}

static void openvox_print_license(int fd)
{
	struct opxtc_license_info license_info;

	int ret = opxtc_module_get_license(&license_info);
	if (ret) {
		ast_cli(fd, "License info: get license error!\n");
	}
	else {
		if (0xffffffff == license_info.max)
			ast_cli(fd, "License info: max=0x%x, current=%u.\n", license_info.max, license_info.current);
		else
			ast_cli(fd, "License info: max=%u, current=%u.\n", license_info.max, license_info.current);
	}
}

static void openvox_print_rtcp(int fd)
{
	if (0 == g_rtcp_enable) {
		ast_cli(fd, "Rtcp is off.\n");
	}
	else {
		ast_cli(fd, "Rtcp if on.\n");
	}
}

static void openvox_set_rtcp(int fd, const char *set)
{
	if (!strcasecmp(set, "on")) {
		g_rtcp_enable = 1;
	}
	else if (!strcasecmp(set, "off")) {
		g_rtcp_enable = 0;
	}
	ast_cli(fd, "Set rtcp to %s.\n", ((g_rtcp_enable == 1) ? "on" : "off"));
}

#if ASTERISK_VERSION_NUM >= 10600
static char *handle_cli_openvox_show_translators(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "openvox show translators";
		e->usage = g_cli_openvox_show_translators_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}
	openvox_print_translators(a->fd);
	return CLI_SUCCESS;
}

static char *handle_cli_openvox_show_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "openvox show transcoding sessions";
		e->usage = g_cli_openvox_show_transcoding_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
	openvox_print_transcoding_sessions(a->fd);
	return CLI_SUCCESS;
}

static char *handle_cli_openvox_flush_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int sessid = -1;
	switch (cmd) {
	case CLI_INIT:
		e->command = "openvox flush transcoding";
		e->usage = g_cli_openvox_flush_transcoding_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc < 3) {
		return CLI_SHOWUSAGE;
	}
	if (a->argc == 4) {
		sessid = atoi(a->argv[3]);
	}
	openvox_flush_transcoding_sessions(a->fd, sessid);
	return CLI_SUCCESS;
}

static char *handle_cli_openvox_show_rtp_statistics(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "openvox show rtp statistics";
		e->usage = g_cli_openvox_show_rtp_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
	openvox_print_rtp_statistics(a->fd);
	return CLI_SUCCESS;
}

static char *handle_cli_openvox_show_license(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "openvox show license";
		e->usage = g_cli_openvox_show_license_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}
	openvox_print_license(a->fd);
	return CLI_SUCCESS;
}

static char *handle_cli_openvox_show_rtcp(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "openvox show rtcp";
		e->usage = g_cli_openvox_show_rtcp_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}
	openvox_print_rtcp(a->fd);
	return CLI_SUCCESS;
}

static char *handle_cli_openvox_set_rtcp(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "openvox set rtcp {on|off}";
		e->usage = g_cli_openvox_set_rtcp_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
	
	openvox_set_rtcp(a->fd, a->argv[3]);
	return CLI_SUCCESS;
}
#else
int handle_cli_openvox_show_translators(int fd, int argc, char *argv[])
{
	openvox_print_translators(fd);
	return 0;
}

int handle_cli_openvox_show_transcoding(int fd, int argc, char *argv[])
{
	openvox_print_transcoding_sessions(fd);
	return 0;
}

int handle_cli_openvox_flush_transcoding(int fd, int argc, char *argv[])
{
	int sessid = -1;
	if (argc == 4) {
		sessid = atoi(argv[3]);
	}
	openvox_flush_transcoding_sessions(fd, sessid);
	return 0;
}

int handle_cli_openvox_show_rtp_statistics(int fd, int argc, char *argv[])
{
	openvox_print_rtp_statistics(fd);
	return 0;
}

int handle_cli_openvox_show_license(int fd, int argc, char *argv[])
{
	openvox_print_license(fd);
	return 0;
}

int handle_cli_openvox_show_rtcp(int fd, int argc, char *argv[])
{
	openvox_print_rtcp(fd);
	return 0;
}

int handle_cli_openvox_set_rtcp(int fd, int argc, char *argv[])
{
	openvox_set_rtcp(fd, argv[3]);
	return 0;
}
#endif

static struct ast_frame *openvox_fakesrc_sample(void)
{
	static struct ast_frame f = {
		.frametype = AST_FRAME_VOICE,
		.samples = OPENVOX_DEFAULT_SAMPLES,
		.src = __PRETTY_FUNCTION__
	};
	return &f;
}

/*! \brief enqueue input frame to openvox session */
static int openvox_framein(struct asterisk_transcoding_pvt *pvt, struct ast_frame *f)
{
	int difftime = 0;
	int res = 0;
	struct timeval now = {0, 0};
	struct openvox_transcoding_session *sess = openvox_get_pvt(pvt);

#if ASTERISK_VERSION_NUM < 10800
	if (!f->subclass) {
#else
	if (!f->subclass.codec) {
#endif
		/* we must return a fake frame for calculation purposes, 
		 * may be we should indeed send something to openvox board? */
		sess->fakestate = FAKE_FRAME_REQUESTED;
		pvt->samples = f->samples;
		return 0;
	}
	if (f->frametype != AST_FRAME_VOICE) {
		OPX_SESS_WARN_TX("refusing to write frame of type %d to Openvox transcoder %llu, we only accept voice in format %s\n", 
				f->frametype, sess->sessid, ast_getformatname(1 << pvt->t->srcfmt));
		return -1;
	}
#if ASTERISK_VERSION_NUM < 10800
	if (f->frametype != AST_FRAME_VOICE || f->subclass != (1 << pvt->t->srcfmt)) {
#else
	if (f->frametype != AST_FRAME_VOICE || f->subclass.codec != (1 << pvt->t->srcfmt)) {
#endif
		OPX_SESS_WARN_TX("refusing to write voice frame in format %s to Openvox transcoder %llu, we only accept voice in format %s\n", 
#if ASTERISK_VERSION_NUM < 10800
				ast_getformatname(f->subclass),
#else
				ast_getformatname(f->subclass.codec),
#endif
			       	sess->sessid, ast_getformatname(1 << pvt->t->srcfmt));
		return -1;
	}

	res = ast_mutex_trylock(&sess->txrtp->txlock);
	if (res) {
		ast_log(LOG_ERROR, "Could not acquire tx lock for RTP tx in session %llu: (err %d)\n", sess->sessid, res);
		return -1;
	}

	if (openvox_rtp_write(sess->txrtp->ast_rtp, f)) {
		OPX_SESS_WARN_TX("Failed to write voice frame in format %s of length %d to Openvox transcoder %llu\n", 
#if ASTERISK_VERSION_NUM < 10800
		ast_getformatname(f->subclass),
#else
		ast_getformatname(f->subclass.codec),
#endif
	       	f->datalen, sess->sessid);
		ast_mutex_unlock(&sess->txrtp->txlock);
		return -1;
	}

	ast_mutex_unlock(&sess->txrtp->txlock);

	now = ast_tvnow();
	if (!sess->lasttxtime.tv_sec) {
		sess->lasttxtime = now;
	} else {
		difftime = ast_tvdiff_ms(now, sess->lasttxtime);
		sess->txavgtime = sess->txavgtime ? ((sess->txavgtime + difftime) / 2) : difftime;
		sess->lasttxtime = now;
	}
	sess->txcount++;

	pvt->samples += f->samples;
	return 0;
}

/*! \brief output frames from the openvox session */
static struct ast_frame *openvox_frameout(struct asterisk_transcoding_pvt *pvt)
{
	/* read from the rtp session and put the result in pvt->f */
	int difftime = 0;
	int res = 0;
	struct timeval now = {0, 0};
	struct ast_frame *f = NULL;
	struct openvox_transcoding_session *sess = openvox_get_pvt(pvt);

	if (FAKE_FRAME_REQUESTED == sess->fakestate) {
		sess->fakestate = FAKE_FRAME_PROVIDED;
		pvt->f.frametype = AST_FRAME_VOICE;
#if ASTERISK_VERSION_NUM < 10800
		pvt->f.subclass = 0;
#else
		pvt->f.subclass.codec = 0;
#endif
		pvt->f.samples = OPENVOX_DEFAULT_SAMPLES;
/* if any version previous to 1.6.1 defines data.ptr, then decrease this version num */
#if ASTERISK_VERSION_NUM >= 10601
		pvt->f.data.ptr = NULL;
#else
		pvt->f.data = NULL;
#endif
#if ASTERISK_VERSION_NUM >= 10400
		//ast_set_flag(&pvt->f, AST_FRFLAG_FROM_TRANSLATOR);
#endif
		pvt->f.offset = 0;
		pvt->f.datalen = 0;
		pvt->f.mallocd = 0;
		pvt->samples = 0;
		/*ast_log(LOG_DEBUG, "Returned fake frame from Openvox transcoder\n");*/
		return ast_frisolate(&pvt->f);
		//return &pvt->f;
	} else if (FAKE_FRAME_PROVIDED == sess->fakestate) {
		/*ast_log(LOG_DEBUG, "Terminated fake frame request to Openvox transcoder\n");*/
		sess->fakestate = FAKE_FRAME_TERMINATED;
		return NULL;
	}

	res = ast_mutex_trylock(&sess->rxrtp->rxlock);
	if (res) {
		ast_log(LOG_ERROR, "Could not acquire rx lock for RTP rx in session %llu: (err %d)\n\n", sess->sessid, res);
		return NULL;
	}

	/* if this is the first time we read we must read any previously buffered data in the RTP socket
	 * otherwise we may end up with audio delays because depending on the board configuration it is
	 * possible the board has been sending us audio even if we did not send audio */
	if (!sess->rxcount) {
		do {
			f = openvox_rtp_read(sess->rxrtp->ast_rtp);
			sess->rxcount++;
		} while (f && f->frametype != AST_FRAME_NULL);
	} else {
		f = openvox_rtp_read(sess->rxrtp->ast_rtp);
	}

	/* calculate average input time */
	now = ast_tvnow();
	if (!sess->lastrxcalltime.tv_sec) {
		sess->lastrxcalltime = now;
	} else {
		difftime = ast_tvdiff_ms(now, sess->lastrxcalltime);
		sess->rxcallavgtime = sess->rxcallavgtime ? ((sess->rxcallavgtime + difftime) / 2) : difftime;
		sess->lastrxcalltime = now;
	}
	sess->rxcallcount++;

	ast_mutex_unlock(&sess->rxrtp->rxlock);

	if (!f) {
		OPX_SESS_WARN_RX("ignoring read NULL frame from Openvox transcoder %llu, we only expect voice in format %s\n", 
				sess->sessid, ast_getformatname(1 << pvt->t->dstfmt));
		return NULL;
	}

	if (f->frametype == AST_FRAME_NULL) {
		sess->rxnullcount++;
		return NULL;
	}
	if (f->frametype != AST_FRAME_VOICE) {
		OPX_SESS_WARN_RX("ignoring read frame of type %d from Openvox transcoder %llu, we only expect voice in format %s\n", 
				f->frametype, sess->sessid, ast_getformatname(1 << pvt->t->dstfmt));
		return NULL;
	}
#if ASTERISK_VERSION_NUM < 10800
	if (f->subclass != (1 << pvt->t->dstfmt)) {
#else
	if (f->subclass.codec != (1 << pvt->t->dstfmt)) {
#endif
		OPX_SESS_WARN_RX("ignoring read voice frame in format %s from Openvox transcoder %llu, we only expect voice in format %s rxrtp=%p\n",
#if ASTERISK_VERSION_NUM < 10800
				ast_getformatname(f->subclass),
#else
				ast_getformatname(f->subclass.codec),
#endif
			       	sess->sessid, ast_getformatname(1 << pvt->t->dstfmt), sess->rxrtp);
		return NULL;
	}

	/* calculate average read time */
	now = ast_tvnow();
	if (!sess->lastrxtime.tv_sec) {
		sess->lastrxtime = now;
	} else {
		difftime = ast_tvdiff_ms(now, sess->lastrxtime);
		sess->rxavgtime = sess->rxavgtime ? ((sess->rxavgtime + difftime) / 2) : difftime;
		sess->lastrxtime = now;
	}
	sess->rxcount++;

	/* check sequence */
	if (sess->lastrxseqno >= 0) {
		static int sessid = 0;
		if (sessid == 0) {
			sessid=sess->sessid;
		}
		if (f->seqno > (sess->lastrxseqno+2) ) {
			//ast_log(LOG_WARNING, "[%lli] Got Seq %i Expecing %li\n", sess->sessid, f->seqno, sess->lastrxseqno+1);
			sess->lostrxpackets += f->seqno - sess->lastrxseqno - 1;
		} 
	} 

	if (sess->lostrxpackets == -1) {
		sess->lostrxpackets = 0;
	}

	sess->lastrxseqno = f->seqno; 

	/* update samples */
	pvt->samples -= f->samples;
	return ast_frisolate(f);
}

static int openvox_create_rtp(void *usr_priv, opxtc_codec_request_leg_t *codec_reg_leg, opxtc_codec_reply_leg_t* codec_reply_leg, void **rtp_fd)
{
	void *fd;
	struct openvox_rtp_handle_ast *opx_rtp;
#if ASTERISK_VERSION_NUM < 10800
	struct sockaddr_in localaddr;
	struct sockaddr_in remoteaddr;
	struct in_addr local_ip = { 0 };

	local_ip.s_addr = htonl(codec_reply_leg->host_ip);
#else
	struct ast_sockaddr local_ip = { {0,} };
	struct ast_sockaddr localaddr;
	struct ast_sockaddr remoteaddr;
	struct sockaddr_in *tmpaddr;
	struct in_addr *ip4a;

	local_ip.ss.ss_family = AF_INET;
	local_ip.len = sizeof(struct sockaddr_in);
	ip4a = &((struct sockaddr_in *)&local_ip.ss)->sin_addr;
	ip4a->s_addr = htonl(codec_reply_leg->host_ip);
#endif

	/* ask to create the session with the local IP provided */
#if ASTERISK_VERSION_NUM < 10800
	openvox_rtp_new(fd, local_ip);
#else
	openvox_rtp_new(fd, &local_ip);
#endif
	if (!fd) {
		ast_log(LOG_ERROR, "Failed to create Openvox transcoding RTP stream\n");
		return -1;
	}
	openvox_rtp_get_local_address(fd, &localaddr);

#if ASTERISK_VERSION_NUM < 10800
	codec_reg_leg->host_udp_port = ntohs(localaddr.sin_port);
	
	/* set the socket where we will send the data to and read from */
	remoteaddr.sin_addr.s_addr = htonl(codec_reply_leg->codec_ip);
	remoteaddr.sin_port = htons(codec_reply_leg->codec_udp_port);
#else
	if (AF_INET == localaddr.ss.ss_family) {
		tmpaddr = (struct sockaddr_in *)&localaddr.ss;
		codec_reg_leg->host_udp_port = ntohs(tmpaddr->sin_port);

		/* set the socket where we will send the data to and read from */
		remoteaddr.ss.ss_family = AF_INET;
		remoteaddr.len = sizeof(struct sockaddr_in);
		tmpaddr = (struct sockaddr_in *)&remoteaddr.ss;
		tmpaddr->sin_addr.s_addr = htonl(codec_reply_leg->codec_ip);
		tmpaddr->sin_port = htons(codec_reply_leg->codec_udp_port);
	}
	else {
		ast_log(LOG_ERROR, "Failed to get local address\n");
		openvox_rtp_destroy(fd);
		return -1;
	}
#endif
	openvox_rtp_set_remote_address(fd, &remoteaddr);
 
	ast_log(LOG_DEBUG, "Creating Codec on host (0x%08X/%d)  codec(0x%08X/%d) CODEC ID=%i ms=%i FD=%p \n",
				codec_reg_leg->host_ip, codec_reg_leg->host_udp_port,
				codec_reply_leg->codec_ip, codec_reply_leg->codec_udp_port, 
				codec_reg_leg->codec_id,  codec_reg_leg->ms, fd);

	opx_rtp = ast_calloc(1, sizeof(*opx_rtp));
	if (!opx_rtp) {
		ast_log(LOG_ERROR, "Failed to create Openvox transcoding RTP stream handle\n");
		openvox_rtp_destroy(fd);
		return -1;
	}
	opx_rtp->ast_rtp = fd;
	ast_mutex_init(&opx_rtp->rxlock);
	ast_mutex_init(&opx_rtp->txlock);
	opx_rtp->flags = 0;

	*rtp_fd = opx_rtp; 

	return 0;
}

static int openvox_destroy_rtp(void *usr_priv, void *fd)
{
	struct openvox_rtp_handle_ast *rtp = fd;
	openvox_rtp_destroy(rtp->ast_rtp);
	rtp->ast_rtp = NULL;
	ast_mutex_destroy(&rtp->rxlock);
	ast_mutex_destroy(&rtp->txlock);
	rtp->flags = 0;
	ast_free(rtp);
	return 0;
}

static int ast_format_to_codec_id(int astformat, int *ptime)
{
	int i;
	for (i = 0; g_codec_map[i].codec_id != -1; i++) {
		if (g_codec_map[i].astformat == astformat) {
			*ptime = g_codec_map[i].maxms;
			return g_codec_map[i].codec_id;
		}
	}
	return -1;
}

static void dump_init_cfg(void)
{
	int i, j;
	opxtc_host_nic_vocallo_cfg_t zero_cfg, *pcfg;
	opxtc_eth_mac_t zero_mac, *pmac;

	memset(&zero_cfg, 0, sizeof(zero_cfg));
	memset(&zero_mac, 0, sizeof(zero_mac));

	ast_log(LOG_DEBUG, "***************dump_init_cfg**************\n");
	ast_log(LOG_DEBUG, "host_nic_vocallo_cfg=%u\n", g_init_cfg.host_nic_vocallo_sz);
	for (i = 0; i < OPXTC_MAX_HOST_VOCALLO_NIC; i++)
	{
		pcfg = &g_init_cfg.host_nic_vocallo_cfg[i];
		if (memcmp(pcfg, &zero_cfg, sizeof(zero_cfg)) == 0)
				continue;
		ast_log(LOG_DEBUG, "host_nic_vocallo_cfg[%d] info:\n", i);
		ast_log(LOG_DEBUG, "  vocallo_base_udp_port=%d, vocallo_ip=0x%x, mask=0x%x,\n",
				pcfg->vocallo_base_udp_port, pcfg->vocallo_ip, pcfg->vocallo_ip_netmask);
		ast_log(LOG_DEBUG, "  host_ip=0x%x, mask=0x%x, host_mac=%s.\n", pcfg->host_ip,
				pcfg->host_ip_netmask, pcfg->host_mac.mac_str);
		for (j = 0; j < OPXTC_MAX_MAC_RANGE; j++)
		{
			pmac = &pcfg->vocallo_macs[j];
			if (memcmp(pmac, &zero_mac, sizeof(zero_mac)) == 0)
				continue;
			ast_log(LOG_DEBUG, "  vocallo_macs[%d]=%s,mac=0x%hhx%hhx%hhx%hhx%hhx%hhx.\n", j, pmac->mac_str, pmac->mac[0], pmac->mac[1], pmac->mac[2], pmac->mac[3], pmac->mac[4], pmac->mac[5]);
		}
		ast_log(LOG_DEBUG, "  vocallo_macs_sz=%u, silence_suppression=%hhu.\n",
				pcfg->vocallo_macs_sz, pcfg->silence_suppression);
	}
	ast_log(LOG_DEBUG, "log=%p, create_rtp=%p, destroy_rtp=%p.\n", g_init_cfg.log,
		       	g_init_cfg.create_rtp, g_init_cfg.destroy_rtp);
	ast_log(LOG_DEBUG, "***************dump_init_cfg end**********\n");
}

static void dump_session(const struct openvox_transcoding_session *sess)
{
#if ASTERISK_VERSION_NUM >= 10400
	const opxtc_codec_request_t *prequest = &(sess->codec_request);
	const opxtc_codec_reply_t *preply = &(sess->codec_reply);

	ast_log(LOG_DEBUG, "****************dump_sess*******************\n");
	ast_log(LOG_DEBUG, "owner=%p,fakestate=%d,\n", sess->owner, sess->fakestate);
	ast_log(LOG_DEBUG, "codec_request.usr_priv=%p,rtcp_enable=%hhu,a:codec_id=%u,ms=%u,host_ip=0x%x,host_netmask=0x%x,host_udp_port=%u;b:codec_id=%u,ms=%u,host_ip=0x%x,host_netmask=0x%x,host_udp_port=%u,\n",
		       	prequest->usr_priv, prequest->rtcp_enable, prequest->a.codec_id, prequest->a.ms,
		       	prequest->a.host_ip, prequest->a.host_netmask, prequest->a.host_udp_port,
			prequest->b.codec_id, prequest->b.ms, prequest->b.host_ip,
			prequest->b.host_netmask, prequest->b.host_udp_port);
	ast_log(LOG_DEBUG, "codec_reply.codec_rtp_ctx_idx=%hu,a:codec_ip=0x%x,codec_netmask=0x%x,codec_udp_port=%u,host_ip=0x%x,host_netmask=0x%x,host_udp_port=%u,iana_code=%u;b:codec_ip=0x%x,codec_netmask=0x%x,codec_udp_port=%u,host_ip=0x%x,host_netmask=0x%x,host_udp_port=%u,iana_code=%u;tx_fd=%p,tx_iana=%u,rx_fd=%p,rx_iana=%u,\n",
		       	preply->codec_rtp_ctx_idx, preply->a.codec_ip, preply->a.codec_netmask,
		       	preply->a.codec_udp_port, preply->a.host_ip,
		       	preply->a.host_netmask, preply->a.host_udp_port, preply->a.iana_code,
		       	preply->b.codec_ip, preply->b.codec_netmask, preply->b.codec_udp_port, preply->b.host_ip,
		       	preply->b.host_netmask, preply->b.host_udp_port, preply->b.iana_code,
			preply->tx_fd, preply->tx_iana, preply->rx_fd, preply->rx_iana);
	ast_log(LOG_DEBUG, "attention: omit rtp_stats!\n");
	ast_log(LOG_DEBUG, "txrtp=%p,rxrtp=%p,sessid=%llu,\n", sess->txrtp, sess->rxrtp, sess->sessid);
	ast_log(LOG_DEBUG, "txcount=%llu,rxcount=%llu,rxcallcount=%llu,rxnullcount=%llu,\n", sess->txcount,
			sess->rxcount, sess->rxcallcount, sess->rxnullcount);
	ast_log(LOG_DEBUG, "txavgtime=%hu,rxavgtime=%hu,rxcallavgtime=%hu,\n", sess->txavgtime, sess->rxavgtime,
			sess->rxcallavgtime);
	ast_log(LOG_DEBUG, "lasttxtime.tv_sec=%ld,tv_usec=%ld,\n", sess->lasttxtime.tv_sec,
		       	sess->lasttxtime.tv_usec);
	ast_log(LOG_DEBUG, "lastrxtime.tv_sec=%ld,tv_usec=%ld,\n", sess->lastrxtime.tv_sec,
			sess->lastrxtime.tv_usec);
	ast_log(LOG_DEBUG, "lastrxcalltime.tv_sec=%ld,tv_usec=%ld,\n", sess->lastrxcalltime.tv_sec,
			sess->lastrxcalltime.tv_usec);
	ast_log(LOG_DEBUG, "lastrxseqno=%ld,lostrxpackets=%ld,rxwarn=%hu,txwarn=%hu.\n", sess->lastrxseqno,
			sess->lostrxpackets, sess->rxwarn, sess->txwarn);
	ast_log(LOG_DEBUG, "******************dump_sess end***************\n");
#endif
}

#if 0
static void opxtc_activate_translators(void)
{
	struct openvox_translator *cur;
	ast_log(LOG_DEBUG, "activate all translators\n");
	AST_LIST_LOCK(&g_openvox_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_openvox_translators, cur, entry) {
		ast_translator_activate(&cur->t);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_openvox_translators);
}

static void opxtc_deactive_translators(void)
{
	struct openvox_translator *cur;
	ast_log(LOG_DEBUG, "deactivate all translators\n");
	AST_LIST_LOCK(&g_openvox_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_openvox_translators, cur, entry) {
		ast_translator_deactivate(&cur->t);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_openvox_translators);
}

static void print_stack()
{
	void *trace[50];
	char **symbols;
	int rc, i;

	rc = backtrace(trace, sizeof(trace)/sizeof(trace[0]));
	symbols = backtrace_symbols(trace, rc);
	for (i = 0; i < rc; i++) {
		ast_log(LOG_DEBUG, "%s\n", symbols[i]);
	}
	free(symbols);
}
#endif

/* \brief initialize the new openvox openvox transcoding session */
static int openvox_new(struct asterisk_transcoding_pvt *pvt)
{
	int err = 0;
	int codec_id = -1;
	int ptime = 0;
	struct openvox_transcoding_session *sess = openvox_get_pvt(pvt);

	/* set a reference to our owner */
	memset(sess, 0, sizeof(*sess));
	sess->owner = pvt;
	sess->lostrxpackets = -1;
	sess->lastrxseqno = -1;

	sess->codec_request.usr_priv = sess;
	sess->codec_request.rtcp_enable = g_rtcp_enable;

#if ASTERISK_VERSION_NUM < 10800
	ast_log(LOG_DEBUG, "Got codec request from Asterisk core from %s (%d) to %s (%d)\n", 
			ast_getformatname(1 << pvt->t->srcfmt), pvt->t->srcfmt, ast_getformatname(1 << pvt->t->dstfmt), pvt->t->dstfmt);
#else
	ast_log(LOG_DEBUG, "Got codec request from Asterisk core from %s (%lld) to %s (%lld)\n", 
			ast_getformatname(1 << pvt->t->srcfmt), pvt->t->srcfmt, ast_getformatname(1 << pvt->t->dstfmt), pvt->t->dstfmt);
#endif

	/*print_stack();*/

	codec_id = ast_format_to_codec_id((1 << pvt->t->srcfmt), &ptime);
	if (codec_id == -1) {
		ast_log(LOG_ERROR, "Failed to map source ast format %d to codec id\n", (1 << pvt->t->srcfmt));
		return -1;
	}
	sess->codec_request.a.codec_id = codec_id;
	sess->codec_request.a.ms = ptime;

	codec_id = ast_format_to_codec_id((1 << pvt->t->dstfmt), &ptime);
	if (codec_id == -1) {
		ast_log(LOG_ERROR, "Failed to map source ast format %d to codec id\n", (1 << pvt->t->dstfmt));
		return -1;
	}
	sess->codec_request.b.codec_id = codec_id;
	sess->codec_request.b.ms = ptime;


#if 0
	ast_log(LOG_DEBUG, "Creating Codec A-CODEC_ID=%i  Codec B-CODEC_ID=%i\n",
				sess->codec_request.a.codec_id, sess->codec_request.b.codec_id);
#else
	ast_log(LOG_DEBUG, "Creating Codec SESSID=%llu A-CODEC_ID/IANA=%i/%i  Codec B-CODEC_ID/IANA=%i/%i  Tx=%p  Rx=%p\n",
				sess->sessid,
				sess->codec_request.a.codec_id, 
				sess->codec_reply.a.iana_code,
				sess->codec_request.b.codec_id,
				sess->codec_reply.b.iana_code,
				sess->txrtp,
				sess->rxrtp);
#endif

	dump_session(sess);
	/* we re-use the global sessions lock to lock the session request in the library */
	AST_LIST_LOCK(&g_openvox_sessions);

	/* create the actual new session */
	err = opxtc_create_transcoding_session(&sess->codec_request, &sess->codec_reply);
	if (err) {
		AST_LIST_UNLOCK(&g_openvox_sessions);
		ast_log(LOG_ERROR, "Failed to create Openvox transcoding session\n");
#if 0
		if (-2 == err) {
			ast_log(LOG_ERROR, "Deactive Openvox translators!\n");
			opxtc_deactive_translators();
			active = 0;
		}
#endif
		sess->rxrtp = NULL;
		sess->txrtp = NULL;
		return -1;
	}

	/* add the new session to the global list for usage statistics */
	sess->txrtp = sess->codec_reply.tx_fd;
	sess->rxrtp = sess->codec_reply.rx_fd;
	ast_log(LOG_DEBUG, "index=%hu, rxrtp->flags=%u, txrtp->flags=%u.\n", sess->codec_reply.codec_rtp_ctx_idx,
			sess->rxrtp->flags, sess->txrtp->flags);

	/* sanity check flags, we should not get RTP already used, if this fail is a bug in the library */
	if (ast_test_flag(sess->txrtp, OPX_RTP_TX)) {
		ast_log(LOG_ERROR, "Invalid transcoding session created! tx rtp it's already owned for transmission by someone else!\n");
		sess->rxrtp = NULL;
		sess->txrtp = NULL;
		if (opxtc_destroy_transcoding_session(sess->codec_reply.codec_rtp_ctx_idx)) {
			ast_log(LOG_ERROR, "Failed to destroy Openvox transcoding session properly (after discovering TX is already in use)\n");
		}
		AST_LIST_UNLOCK(&g_openvox_sessions);
		return -1;
	}

	if (ast_test_flag(sess->rxrtp, OPX_RTP_RX)) {
		ast_log(LOG_ERROR, "Invalid transcoding session created! rx rtp it's already owned for transmission by someone else!\n");
		sess->rxrtp = NULL;
		sess->txrtp = NULL;
		if (opxtc_destroy_transcoding_session(sess->codec_reply.codec_rtp_ctx_idx)) {
			ast_log(LOG_ERROR, "Failed to destroy Openvox transcoding session properly (after discovering RX is already in use)\n");
		}
		AST_LIST_UNLOCK(&g_openvox_sessions);
		return -1;
	}

	ast_set_flag(sess->txrtp, OPX_RTP_TX);
	ast_set_flag(sess->rxrtp, OPX_RTP_RX);
	ast_log(LOG_DEBUG, "after set, index=%hu, rxrtp->flags=%u, txrtp->flags=%u.\n", sess->codec_reply.codec_rtp_ctx_idx,
			sess->rxrtp->flags, sess->txrtp->flags);

	sess->sessid = g_next_session_id++;
	g_total_sessions++;
	
	ast_log(LOG_DEBUG, "Created Codec SESSID=%llu A-CODEC_ID/IANA=%i/%i  Codec B-CODEC_ID/IANA=%i/%i  Tx=%p  Rx=%p\n",
				sess->sessid,
				sess->codec_request.a.codec_id, 
				sess->codec_reply.a.iana_code,
				sess->codec_request.b.codec_id,
				sess->codec_reply.b.iana_code,
				sess->txrtp,
				sess->rxrtp);

	AST_LIST_INSERT_HEAD(&g_openvox_sessions, sess, entry);

	AST_LIST_UNLOCK(&g_openvox_sessions);

	ast_log(LOG_DEBUG, "Openvox Transcoding Session %llu Created\n", sess->sessid);

	ast_log(LOG_DEBUG, "Openvox Transcoding Session %llu Rx %s in %d.%d.%d.%d:%d from board at %d.%d.%d.%d:%d fd=%p\n",
			sess->sessid, ast_getformatname(1 << sess->owner->t->dstfmt), 
			OPXTC_NIPV4(sess->codec_reply.b.host_ip), sess->codec_reply.b.host_udp_port,
			OPXTC_NIPV4(sess->codec_reply.b.codec_ip), sess->codec_reply.b.codec_udp_port, sess->rxrtp); 

	ast_log(LOG_DEBUG, "Openvox Transcoding Session %llu Tx %s from %d.%d.%d.%d:%d to board at %d.%d.%d.%d:%d fd=%p\n",
			sess->sessid, ast_getformatname(1 << sess->owner->t->srcfmt),
			OPXTC_NIPV4(sess->codec_reply.a.host_ip), sess->codec_reply.a.host_udp_port,
			OPXTC_NIPV4(sess->codec_reply.a.codec_ip), sess->codec_reply.a.codec_udp_port, sess->txrtp);
	dump_session(sess);
	return 0;
}

static void openvox_destroy(struct asterisk_transcoding_pvt *pvt)
{
	struct openvox_transcoding_session *sess = openvox_get_pvt(pvt);

	ast_log(LOG_DEBUG, "Destroying Openvox transcoding session %llu\n", sess->sessid);

	AST_LIST_LOCK(&g_openvox_sessions);
	
	/* clear sanity flags, we're no longer using this RTP session for tx/rx */
	ast_clear_flag(sess->txrtp, OPX_RTP_TX);
	ast_clear_flag(sess->rxrtp, OPX_RTP_RX);

	dump_session(sess);
	if (opxtc_destroy_transcoding_session(sess->codec_reply.codec_rtp_ctx_idx)) {
		ast_log(LOG_ERROR, "Failed to destroy Openvox transcoding session %llu properly\n", sess->sessid);
	}

	AST_LIST_REMOVE(&g_openvox_sessions, sess, entry);
	g_total_sessions--;
	AST_LIST_UNLOCK(&g_openvox_sessions);
	dump_session(sess);

#if 0
	if (!active) {
		ast_log(LOG_ERROR, "Activate Openvox translators!\n");
		opxtc_activate_translators();
		active = 1;
	}
#endif
}

#if ASTERISK_VERSION_NUM < 10400
#define LOCAL_COUNT_INC  \
	ast_mutex_lock(&usemutex); \
	localusecnt++; \
	ast_mutex_unlock(&usemutex); 
#define LOCAL_COUNT_DEC  \
	ast_mutex_lock(&usemutex); \
	localusecnt--; \
	ast_mutex_unlock(&usemutex); 
static struct asterisk_transcoding_pvt *openvox_new_g729toulaw()
{
	struct asterisk_transcoding_pvt *pvt = NULL;
	LOCAL_COUNT_INC;
	pvt = calloc(1, sizeof(struct asterisk_transcoding_pvt));
	if (!pvt) {
		ast_log(LOG_ERROR, "Could not allocate pvt structure\n");
		LOCAL_COUNT_DEC;
		return NULL;
	}
	pvt->dstfmt = 2;
	pvt->srcfmt = 8;
	pvt->t = pvt;
	if (openvox_new(pvt)) {
		free(pvt);
		LOCAL_COUNT_DEC;
		return NULL;
	}
	return pvt;
}

static struct asterisk_transcoding_pvt *openvox_new_ulawtog729()
{
	struct asterisk_transcoding_pvt *pvt = NULL;
	LOCAL_COUNT_INC;
	pvt = calloc(1, sizeof(struct asterisk_transcoding_pvt));
	if (!pvt) {
		LOCAL_COUNT_DEC;
		ast_log(LOG_ERROR, "Could not allocate pvt structure\n");
		return NULL;
	}
	pvt->dstfmt = 8;
	pvt->srcfmt = 2;
	pvt->t = pvt;
	if (openvox_new(pvt)) {
		free(pvt);
		LOCAL_COUNT_DEC;
		return NULL;
	}
	return pvt;
}

static void openvox_destroy_pvt_old(struct asterisk_transcoding_pvt *pvt)
{
	openvox_destroy(pvt);
	free(pvt);
	LOCAL_COUNT_DEC;
}
#endif /* endif ASTERISK 1.2 */

static int load_nic_network_information(char *nic, opxtc_host_nic_vocallo_cfg_t *cfg)
{
	struct ifreq ifr;
	int sock;
	char *mac;
	int j = 0, k = 0, ret = 0;
	
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (-1 == sock) {
		ast_log(LOG_ERROR, "Failed to create socket to query network configuration for NIC %s: %s\n", nic, strerror(errno));
		return -1;
	}

	strncpy(ifr.ifr_name, nic, sizeof(ifr.ifr_name)-1);
	ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';

	if (-1 == ioctl(sock, SIOCGIFADDR, &ifr)) {
		ast_log(LOG_ERROR, "Failed to query IP address for NIC %s: %s\n", nic, strerror(errno));
		ret = -1;
		goto done;
	}
	cfg->host_ip = ntohl(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr);

	if (-1 == ioctl(sock, SIOCGIFNETMASK, &ifr)) {
		ast_log(LOG_ERROR, "Failed to query network address for NIC %s: %s\n", nic, strerror(errno));
		ret = -1;
		goto done;
	}
	cfg->host_ip_netmask = ntohl(((struct sockaddr_in *)&ifr.ifr_netmask)->sin_addr.s_addr);

	if (-1 == ioctl(sock, SIOCGIFHWADDR, &ifr)) {
		ast_log(LOG_ERROR, "Failed to query HW address for NIC %s: %s\n", nic, strerror(errno));
		ret = -1;
		goto done;
	}

	mac = cfg->host_mac.mac_str;
	for (j = 0, k = 0; j < 6; j++) {
		k += snprintf(mac + k, sizeof(cfg->host_mac.mac_str) - k, j ? "-%02X" : "%02X",
				(int)(unsigned int)(unsigned char)ifr.ifr_hwaddr.sa_data[j]);
		cfg->host_mac.mac[j] = ifr.ifr_hwaddr.sa_data[j];
	}
	mac[sizeof(cfg->host_mac.mac_str)-1] = '\0';

done:
	close(sock);

	return ret;
}

/* updates the list of codecs and return the number of codecs/translators added */
static unsigned int update_codec_list(char codec_list[MAX_CODECS_LIST][MAX_CODEC_STR], size_t size, int index, const char *value)
{
	unsigned int i = index;
	char line[1024];
	char *c = NULL;
	char *tok = NULL;

	ast_log(LOG_DEBUG, "Parsing codec string %s\n", value);
	ast_copy_string(line, value, sizeof(line));

	/* the value string is in format codec,codec,codec where codec can either be a codec name 
	 * or a translator name (translator names are codec1tocodec2 names ) */
	tok = strtok_r(line, ",", &c);
	if (tok == NULL) {
		ast_log(LOG_WARNING, "not found any codec specification!\n");
		return 0;
	}

	ast_copy_string(codec_list[i], tok, sizeof(codec_list[0]));
	ast_log(LOG_NOTICE, "found translator %s\n", codec_list[i]);
	i++;

	while ((tok = strtok_r(NULL, ",", &c))) {
		ast_copy_string(codec_list[i], tok, sizeof(codec_list[0]));
		ast_log(LOG_NOTICE, "found translator %s\n", codec_list[i]);
		i++;
	}

	return (i - index);
}

static int openvox_parse_config(int reload)
{
	struct ast_variable *v = NULL;
	char *category = NULL;
	int nic_idx = 0;
	char ipbuf1[255];
	char ipbuf2[255];
	const char *ipaddr = NULL;
	struct sockaddr_in vocalloaddr;
	struct in_addr print_host_ip;
	struct in_addr print_host_netmask;
#if ASTERISK_VERSION_NUM >= 10600
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct ast_config *cfg = ast_config_load(OPENVOX_TRANSCODE_CONFIG, config_flags);
	if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_verbose("File " OPENVOX_TRANSCODE_CONFIG " was not changed, ignoring reload\n");
		return 0;
	}
#else
	struct ast_config *cfg = ast_config_load(OPENVOX_TRANSCODE_CONFIG);
#endif

#if ASTERISK_VERSION_NUM > 10602
	if (cfg == CONFIG_STATUS_FILEMISSING || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_WARNING, "Openvox transcoding configuration file " OPENVOX_TRANSCODE_CONFIG " is missing or invalid\n");
		return -1;
	}
#endif
	if (!cfg) {
		ast_log(LOG_WARNING, "Failed to Openvox transcoding configuration file " OPENVOX_TRANSCODE_CONFIG ", make sure the file exists and is readable\n");
		return -1;
	}

	memset(&g_init_cfg, 0, sizeof(g_init_cfg));
	memset(&vocalloaddr, 0, sizeof(vocalloaddr));
	memset(g_codec_register_list, 0, sizeof(g_codec_register_list));
	memset(g_codec_noregister_list, 0, sizeof(g_codec_noregister_list));
	g_codec_register_list_count = 0;
	while ((category = ast_category_browse(cfg, category))) {

		if (!strcasecmp("general", category)) {
			for (v = ast_variable_browse(cfg, category); v; v = v->next) {
				if (!strcasecmp(v->name, "unregisterconflictingcodecs")) {
					g_attempt_unload_conflicting_translators = ast_true(v->value);
					if (g_attempt_unload_conflicting_translators) {
						ast_log(LOG_NOTICE, "unregisterconflictingcodecs is now enabled, this may cause warnings when unloading other codec modules\n");
					}
				} else if (!strcasecmp(v->name, "allownondahditiming")) {
					g_allow_nondahdi_timing = ast_true(v->value);
					if (g_allow_nondahdi_timing) {
						ast_log(LOG_WARNING, "BE AWARE THAT NON-DAHDI TIMING HAS BEEN KNOWN TO CAUSE AUDIO LOSS AND/OR CRASHES\n");
						ast_log(LOG_WARNING, "BE AWARE THAT NON-DAHDI TIMING HAS BEEN KNOWN TO CAUSE AUDIO LOSS AND/OR CRASHES\n");
						ast_log(LOG_WARNING, "BE AWARE THAT NON-DAHDI TIMING HAS BEEN KNOWN TO CAUSE AUDIO LOSS AND/OR CRASHES\n");
					}
				} else if (!strcasecmp(v->name, "register")) {
					g_codec_register_list_count += update_codec_list(g_codec_register_list, ARRAY_LEN(g_codec_register_list), g_codec_register_list_count, v->value);
				} else if (!strcasecmp(v->name, "noregister")) {
					g_codec_noregister_list_count += update_codec_list(g_codec_noregister_list, ARRAY_LEN(g_codec_noregister_list), g_codec_noregister_list_count, v->value);
				} else {
					ast_log(LOG_WARNING, "Ignored unknown Openvox codec setting %s in [general]\n", v->name);
				}
			}
			continue;
		}

		if (load_nic_network_information(category, &g_init_cfg.host_nic_vocallo_cfg[nic_idx])) {
			ast_log(LOG_WARNING, "Ignoring NIC %s, failed to retrieve IP network configuration\n", category);
			continue;
		}
		ast_log(LOG_DEBUG, "Configuring NIC %s ( %s )\n", category, g_init_cfg.host_nic_vocallo_cfg[nic_idx].host_mac.mac_str);

		/* set defaults in case is some parameters are not specified */
		g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_base_udp_port = OPENVOX_DEFAULT_UDP_PORT;
		g_init_cfg.host_nic_vocallo_cfg[nic_idx].silence_suppression = 0;

		for (v = ast_variable_browse(cfg, category); v; v = v->next) {
			if (!strcasecmp(v->name, "baseudp")) {
				g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_base_udp_port = atoi(v->value);
				ast_log(LOG_DEBUG, "Using base udp %s (%d)\n", v->value, g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_base_udp_port);
			}
			else if (!strcasecmp(v->name, "vocalloaddr")) {
#if ASTERISK_VERSION_NUM >= 10600
				if (ast_parse_arg(v->value, PARSE_INADDR, &vocalloaddr)) {
					ast_log(LOG_WARNING, "Invalid Openvox transcoding vocallo addr: %s\n", v->value);
					vocalloaddr.sin_addr.s_addr = 0;
					continue;
				}
#else
				struct ast_hostent ahp;
				struct hostent *hp;
				if (!(hp = ast_gethostbyname(v->value, &ahp))) {
					ast_log(LOG_WARNING, "Invalid Openvox transcoding Vocallo host, could not resolve: %s\n", v->value);
					continue;
				} else {
					memcpy(&vocalloaddr.sin_addr, hp->h_addr, sizeof(vocalloaddr.sin_addr));
				}
#endif
				g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_ip = ntohl(vocalloaddr.sin_addr.s_addr);
				ast_log(LOG_DEBUG, "Using transcoding vocallo host %s (0x%08X)\n", v->value, 
						g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_ip);
			} else if (!strcasecmp(v->name, "silence")) {
				g_init_cfg.host_nic_vocallo_cfg[nic_idx].silence_suppression = ast_true(v->value) ? 1 : 0;
			} else {
				ast_log(LOG_WARNING, "Ignored unknown Openvox codec setting %s in NIC %s configuration\n", v->name, category);
			}
		}

		/* validate some stuff */
		if (!g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_ip) {
			ast_log(LOG_ERROR, "Ignoring Openvox transcoding interface %s, you must specify a valid vocalloaddr parameter\n", category);
			continue;
		}

		print_host_ip.s_addr = htonl(g_init_cfg.host_nic_vocallo_cfg[nic_idx].host_ip);
		print_host_netmask.s_addr = htonl(g_init_cfg.host_nic_vocallo_cfg[nic_idx].host_ip_netmask);
#if ASTERISK_VERSION_NUM >= 10400
		ipaddr = ast_inet_ntoa(print_host_ip);
		strncpy(ipbuf1, ipaddr, sizeof(ipbuf1)-1);
		ipbuf1[sizeof(ipbuf1)-1] = 0;

		ipaddr = ast_inet_ntoa(print_host_netmask);
		strncpy(ipbuf2, ipaddr, sizeof(ipbuf2)-1);
		ipbuf1[sizeof(ipbuf2)-1] = 0;
		ast_verbose("Openvox transcoding interface %s using IP address %s, netmask %s, silence suppression = %s\n", 
				category, ipbuf1, ipbuf2, g_init_cfg.host_nic_vocallo_cfg[nic_idx].silence_suppression ? "yes" : "no");
#else
		ast_verbose("Openvox transcoding interface %s using IP address %s, netmask %s\n", 
				category, ast_inet_ntoa(buf1, sizeof(ipbuf1), print_host_ip), 
				ast_inet_ntoa(buf2, sizeof(ipbuf2), print_host_netmask));
#endif
		nic_idx++;
	}

	ast_config_destroy(cfg);

	g_init_cfg.host_nic_vocallo_sz = nic_idx;

	if (!g_codec_register_list_count) {
		ast_log(LOG_NOTICE, "No 'register' codecs where specified in configuration file, defaulting to g729, g726 and g722\n");
		ast_copy_string(g_codec_register_list[0], "g729", sizeof(g_codec_register_list[0]));
		ast_copy_string(g_codec_register_list[1], "g726", sizeof(g_codec_register_list[0]));
		ast_copy_string(g_codec_register_list[2], "g722", sizeof(g_codec_register_list[0]));
		g_codec_register_list_count = 3;
	}
	return 0;
}

/*! \brief standard module stuff */

#if ASTERISK_VERSION_NUM >= 10400
static 
#endif
int reload(void)
{
	if (openvox_parse_config(1)) {
		return AST_MODULE_LOAD_DECLINE;
	}
	return AST_MODULE_LOAD_SUCCESS;
}

static void unregister_translators(void)
{
	struct openvox_translator *cur;
	ast_log(LOG_DEBUG, "Unregistering all translators\n");
	AST_LIST_LOCK(&g_openvox_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_openvox_translators, cur, entry) {
		ast_unregister_translator(&cur->t);
		ast_free(cur);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_openvox_translators);
}

#if ASTERISK_VERSION_NUM >= 10400
static 
#endif
int unload_module(void)
{
	ast_cli_unregister_multiple(g_openvox_cli, ARRAY_LEN(g_openvox_cli));
	unregister_translators();
	opxtc_deactivate_modules();
	if (g_sched) {
		sched_context_destroy(g_sched);
		g_sched = NULL;
	}
	return 0;
}

static int openvox_logger(int level, char *fmt, ...)
{
	char *data;
	int ret = 0;
	va_list ap;

	va_start(ap, fmt);
	ret = ast_vasprintf(&data, fmt, ap);
	if (ret == -1) {
		return ret;
	}
	va_end(ap);

	switch (level) {
	case OPXTC_LOGLEVEL_DEBUG:
		ast_log(LOG_DEBUG, "%s\n", data);  
		break;
	case OPXTC_LOGLEVEL_WARN:
		ast_log(LOG_WARNING, "%s\n", data);  
		break;
	case OPXTC_LOGLEVEL_INFO:
		ast_log(LOG_NOTICE,"%s\n", data);
		break;
	case OPXTC_LOGLEVEL_STATS:
		break;
	case OPXTC_LOGLEVEL_ERROR:
		ast_log(LOG_ERROR,"%s\n", data);
		break;
	case OPXTC_LOGLEVEL_CRIT:
		ast_log(LOG_ERROR,"%s\n", data);
		break;
	default:
		ast_log(LOG_ERROR,"Error: invalid loglevel %i\n",level);
		return -1;
	}

	ast_free(data);
	return 0;
}

static int register_translators(void)
{
	int ret = 0;
	struct openvox_translator *cur;
	ast_log(LOG_DEBUG, "Registering all translators\n");
	AST_LIST_LOCK(&g_openvox_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_openvox_translators, cur, entry) {
		if (ast_register_translator(&cur->t)) {
			ast_log(LOG_ERROR, "Failed to register translator %s\n", cur->t.name);
			ret = -1;
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_openvox_translators);
	return ret;
}

static void attempt_unload_conflicting_translators(int srcfmt, int dstfmt)
{
	int steps = 0;
	struct ast_trans_pvt *path;
	char cname[50];
retryunload:
	steps = ast_translate_path_steps(dstfmt, srcfmt);
	if (steps == 1) {
		path = ast_translator_build_path(dstfmt, srcfmt);
		if (!path) {
			ast_log(LOG_ERROR, "Wow, steps from %d(%s) to %d(%s) but no translation path\n",
				srcfmt, ast_getformatname(srcfmt), dstfmt, ast_getformatname(dstfmt));
			return;
		}
		/* this is evil and will cause the owner modules for this translator to 
		 * fail to unregister it but is better than crashing */
		ast_copy_string(cname, path->t->name, sizeof(cname));
		if (ast_unregister_translator(path->t)) {
			ast_log(LOG_ERROR, "Failed to force unregistration of codec %s, crash may occur!", cname);
		} else {
			ast_log(LOG_NOTICE, "Forced unregistration of codec %s\n", cname);
		}
		ast_translator_free_path(path);
		goto retryunload;
	}
}

static int in_list(char codec_list[MAX_CODECS_LIST][MAX_CODEC_STR], size_t size, char *value, int explicit)
{
	int i;
	for (i = 0; i < size; i++) {
		if (!strcasecmp(value, codec_list[i])) {
			return 1;
		}
		/* if any list element has "all" and explicit search has not been requested 
		   then accept the value */
		if (!explicit && !strcasecmp(codec_list[i], "all")) {
			return 1;
		}
	}
	return 0;
}

static int print_list(char codec_list[MAX_CODECS_LIST][MAX_CODEC_STR], size_t size, const char *name)
{
	int i;
	ast_log(LOG_DEBUG, "%s list has %zd elements: \n", name, size);
	for (i = 0; i < size; i++) {
		ast_log(LOG_DEBUG, "%s\n", codec_list[i]);
	}
	return 0;
}

static int add_translator(int srcfmt, int dstfmt)
{
	struct openvox_translator *vt;
	char tcname[255];

	snprintf(tcname, sizeof(tcname), "%sto%s", ast_getformatname(srcfmt), ast_getformatname(dstfmt));
	if (!in_list(g_codec_register_list, g_codec_register_list_count, ast_getformatname(srcfmt), 0) &&
	    !in_list(g_codec_register_list, g_codec_register_list_count, ast_getformatname(dstfmt), 0) &&
	    !in_list(g_codec_register_list, g_codec_register_list_count, tcname, 0)
	   ) {
		/* neither srcfmt, dstfmt or the full translator names were found in the register list */
		ast_log(LOG_DEBUG, "Not registering translator %s because was not found in the register list\n", tcname);
		return 0;
	}
	if (in_list(g_codec_noregister_list, g_codec_noregister_list_count, ast_getformatname(srcfmt), 1) ||
	    in_list(g_codec_noregister_list, g_codec_noregister_list_count, ast_getformatname(dstfmt), 1) ) {
		/* if the src or dest formats are in the noregister, the only way to proceeed is that 
		 * the explicit codec is in the register list */
		if (!in_list(g_codec_register_list, g_codec_register_list_count, tcname, 1)) {
			ast_log(LOG_DEBUG, "Not registering translator %s because it's src/dst was found "
					"in the noregister list and was not found explicitly in the register list\n", tcname);
			return 0;
		}
	}
		    
	if (in_list(g_codec_noregister_list, g_codec_noregister_list_count, tcname, 1)) {
		/* neither srcfmt, dstfmt or the full translator names were found in the register list */
		ast_log(LOG_DEBUG, "Not registering translator %s because it was found explicitly in the noregister list\n", tcname);
		return 0;
	}

	if (!(vt = ast_calloc(1, sizeof(*vt)))) {
		return -1;
	}

	vt->t.srcfmt = srcfmt;
	vt->t.dstfmt = dstfmt;
	vt->t.framein = openvox_framein;
	vt->t.frameout = openvox_frameout;
	vt->t.sample = openvox_fakesrc_sample;
#if ASTERISK_VERSION_NUM >= 10400
	vt->t.desc_size = sizeof(struct openvox_transcoding_session);
	vt->t.newpvt = openvox_new;
	vt->t.destroy = openvox_destroy;
	vt->t.buf_size = OPENVOX_TRANSLATOR_BUFFER_SIZE;
#else
#error "FIXME: Asterisk 1.2 not supported yet"
	vt->t.newpvt = i == 0 ? openvox_new_g729toulaw : openvox_new_ulawtog729;
	vt->t.destroy = openvox_destroy_pvt_old;
#endif
	snprintf((char *)vt->t.name, sizeof(vt->t.name), "opx%s", tcname);

	if (g_attempt_unload_conflicting_translators) {
		attempt_unload_conflicting_translators(srcfmt, dstfmt);
	}

	/* it is important to NOT register the translator just yet, just add it to our linked list 
	 * reason being that attempt_unload_conflicting_translators() has better chances to succed
	 * if we dont register our translations in the mix */
	AST_LIST_LOCK(&g_openvox_translators);
	AST_LIST_INSERT_HEAD(&g_openvox_translators, vt, entry);
	AST_LIST_UNLOCK(&g_openvox_translators);

	return 0;
}

static int entry_callback(const char *module, const char *desc, int usecnt, const char *like)
{
	if (strcasestr(module, "res_timing_dahdi")) {
		g_dahdi_timing = 1;
		return 1;
	}

	if (strcasestr(module, like)) {
		return 1;
	}

	return 0;
}

static int check_dahdi_timing(void)
{
#if ASTERISK_VERSION_NUM > 10602
	int timing_sources = 0;
	if (g_allow_nondahdi_timing) {
		return 0;
	}
	timing_sources = ast_update_module_list(entry_callback, "res_timing_");
	if (timing_sources != 1 || !g_dahdi_timing) {
		return -1;
	}
#endif
	if (0) {
		entry_callback(NULL, NULL, 0, NULL);
	}
	return 0;
}

#if ASTERISK_VERSION_NUM >= 10400
static 
#endif
int load_module(void)
{
	void *fakertp = NULL;
	int i, j;
	int detected = 0, activated = 0;
	uid_t procuid;
	int64_t astiana;
#if ASTERISK_VERSION_NUM < 10800
	struct in_addr localhost_ip = { 0 };
#else
	struct ast_sockaddr localhost_ip = { {0,} };
	struct in_addr *ip4a;
#endif

	/* this module can only be run as root */
	procuid = geteuid();
	if (procuid != 0) {
		ast_log(LOG_ERROR, "This module requires to create raw sockets and only root can do that. Please run Asterisk as root\n");
		return AST_MODULE_LOAD_FAILURE;
	}

	dump_init_cfg();
	if (openvox_parse_config(0)) {
		return AST_MODULE_LOAD_FAILURE;
	}

	dump_init_cfg();
	ast_log(LOG_DEBUG, "before call check_dahdi_timing().\n");
	if (check_dahdi_timing()) {
		ast_log(LOG_ERROR, "Refusing to load this module with non-dahdi timing, please disable all res_timing_*.so modules and leave res_timing_dahdi.so as the only timing source in modules.conf\n");
		ast_log(LOG_ERROR, "You can load this module with other timing sources by setting allownondahditiming=yes in %s general configuration section\n", OPENVOX_TRANSCODE_CONFIG);
		return AST_MODULE_LOAD_FAILURE;
	}

	ast_log(LOG_DEBUG, "before call sched_context_create().\n");
	g_init_cfg.log = openvox_logger;
	g_init_cfg.create_rtp = openvox_create_rtp;
	g_init_cfg.destroy_rtp = openvox_destroy_rtp;

	dump_init_cfg();
	if (!(g_sched = sched_context_create())) {
		ast_log(LOG_ERROR, "Unable to create scheduler context\n");
		return AST_MODULE_LOAD_FAILURE;
	}

	dump_init_cfg();
	ast_log(LOG_DEBUG, "before call opxtc_deactivate_modules().\n");
	opxtc_deactivate_modules();

	dump_init_cfg();
	ast_log(LOG_DEBUG, "before call opxtc_detect_init_modules().\n");
	if (opxtc_detect_init_modules(&g_init_cfg, &detected)) {
		ast_log(LOG_ERROR, "Failed to detect init openvox transcoding library\n");
		goto failure;
	}
    
	dump_init_cfg();
	ast_log(LOG_DEBUG, "before call opxtc_activate_modules().\n");
	if (opxtc_activate_modules(&g_init_cfg, &activated)) {
		ast_log(LOG_ERROR, "Failed to activate openvox transcoding library\n");
		goto failure;
	}   
	dump_init_cfg();

	/* ask to create the session with the local IP provided */
#if ASTERISK_VERSION_NUM < 10800
	/* call opxtc_set_iana_code_based_on_codec_id to set Asterisk IANA codes (see main/rtp.c) */
	inet_aton("127.0.0.1", &localhost_ip);

	openvox_rtp_new(fakertp, localhost_ip);
#else
	localhost_ip.ss.ss_family = AF_INET;
	localhost_ip.len = sizeof(struct sockaddr_in);
	ip4a = &((struct sockaddr_in *)&localhost_ip.ss)->sin_addr;
	/* call opxtc_set_iana_code_based_on_codec_id to set Asterisk IANA codes (see main/rtp.c) */
	inet_aton("127.0.0.1", ip4a);

	openvox_rtp_new(fakertp, &localhost_ip);
#endif
	if (!fakertp) {
		ast_log(LOG_ERROR, "Failed to create fake RTP session on 127.0.0.1 on startup\n");
		goto failure;
	}

	/* add all supported translators */
	ast_log(LOG_DEBUG, "Registering translators\n");
	print_list(g_codec_register_list, g_codec_register_list_count, "register");
	print_list(g_codec_noregister_list, g_codec_noregister_list_count, "noregister");
	for (i = 0; g_codec_map[i].codec_id != -1; i++) {

#if ASTERISK_VERSION_NUM < 10800
		astiana = ast_rtp_lookup_code(fakertp, 1, g_codec_map[i].astformat);
#else
		astiana = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(fakertp), 1, g_codec_map[i].astformat);
#endif
		ast_log(LOG_DEBUG, "Mapping %s to IANA %lld\n", ast_getformatname(g_codec_map[i].astformat), astiana);
		opxtc_set_iana_code_based_on_codec_id(g_codec_map[i].codec_id, astiana);

		for (j = i; g_codec_map[j].codec_id != -1; j++) {

			if (g_codec_map[i].codec_id == g_codec_map[j].codec_id) {
				continue;
			}

			if (add_translator(g_codec_map[i].astformat, g_codec_map[j].astformat)) {
				goto failure;
			}

			if (add_translator(g_codec_map[j].astformat, g_codec_map[i].astformat)) {
				goto failure;
			}
		}
	}

	ast_log(LOG_DEBUG, "before call openvox_rtp_destroy().\n");
	openvox_rtp_destroy(fakertp);

	ast_log(LOG_DEBUG, "before call register_translators().\n");
	/* register all translators with asterisk */
	if (register_translators()) {
		goto failure;
	}

	ast_log(LOG_DEBUG, "before call ast_cli_register_multiple().\n");
	/* register CLI helpers */
	ast_cli_register_multiple(g_openvox_cli, ARRAY_LEN(g_openvox_cli));

	return AST_MODULE_LOAD_SUCCESS;

failure:
	unregister_translators();
	sched_context_destroy(g_sched);
	g_sched = NULL;
	return AST_MODULE_LOAD_FAILURE;
}

#if ASTERISK_VERSION_NUM >= 10400
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, g_openvox_module_description,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
);
#else
char *description(void)
{
	return g_openvox_module_description;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
#endif
